//i7z.c
/* ----------------------------------------------------------------------- *
 *
 *   Copyright 2009 Abhishek Jaiantilal
 *
 *   Under GPL v2
 *
 * ----------------------------------------------------------------------- */

#include <memory.h>
#include <errno.h>
#include <stdio.h>
#include <fcntl.h>
#include <unistd.h>
#include <stdlib.h>
#include <getopt.h>
#include <inttypes.h>
#include <sys/types.h>
#include <sys/time.h>
#include <time.h>
#include <ncurses.h>

#include "i7z.h"
//#include "CPUHeirarchy.h"

int Single_Socket();
int Dual_Socket();

int numPhysicalCores, numLogicalCores;
double TRUE_CPU_FREQ;

#define IA32_THERM_STATUS 0x19C
#define IA32_TEMPERATURE_TARGET 0x1a2

//Info: I start from index 1 when i talk about cores on CPU

//Monitoring Version - 1
//4 events at most can be monitored
//put in the value in PERFEVTSEL and the counts are in PMC
#define IA32_PERFEVTSEL0 0x186
#define IA32_PERFEVTSEL1 0x187
#define IA32_PERFEVTSEL2 0x188
#define IA32_PERFEVTSEL3 0x189
#define IA32_PMC0 0x0C1
#define IA32_PMC1 0x0C2
#define IA32_PMC2 0x0C3
#define IA32_PMC3 0x0C4

//Monitoring Version - 2
//Another 3 events can be then monitored  in IA32_FIXED_CTR0 to CTR2
#define IA32_FIXED_CTR_CTL 0x38D
#define IA32_PERF_GLOBAL_CTRL 0x38F
#define IA32_FIXED_CTR0 0x309 //Instruction_Retired.Any
#define IA32_FIXED_CTR1 0x30A //CPU_CLK_Unhalted.Core
#define IA32_FIXED_CTR2 0x30B //CPU_CLK_Unhalted.Ref

//Performance event number and umask value
#define MEM_INST_RETIRED_LOADS_EVENT 0x0B
#define MEM_INST_RETIRED_LOADS_UMASK 0x01
#define MEM_INST_RETIRED_LOADS 0x010B

#define MEM_INST_RETIRED_STORES_EVENT 0x0B
#define MEM_INST_RETIRED_STORES_UMASK 0x02
#define MEM_INST_RETIRED_STORES 0x020B

#define FP_COMP_OPS_EXE_X87_EVENT 0x10
#define FP_COMP_OPS_EXE_X87_UMASK 0x01
#define FP_COMP_OPS_EXE_X87 0x0110

#define FP_COMP_OPS_EXE_SSE_FP_EVENT 0x10
#define FP_COMP_OPS_EXE_SSE_FP_UMASK 0x04
#define FP_COMP_OPS_EXE_SSE_FP 0x0410


int error_indx;

struct therm_status_struct{
    int reading_valid;
    int resolution_deg_celcius;
    int digital_readout;
    int power_limit_notification_status;
    int power_limit_notification_log;
    int thermal_threshold_2_log;
    int thermal_threshold_2_status;
    int thermal_threshold_1_log;
    int thermal_threshold_1_status;
    int critical_temperature_log;
    int critical_temperature_status;
    int PROCHOT_log;
    int PROCHOT_event;
    int thermal_status_log;
    int thermal_status;
    
    int PROCHOT_temp; //Tj
    int current_core_temp; //PROCHOT_temp - digital_readout
};

void Print_Thermal_Status(struct therm_status_struct therm_status){
    printf("Reading Valid: %d\n",therm_status.reading_valid);
    printf("Deg Celcius: %d\n",therm_status.resolution_deg_celcius);
    printf("Digital Readout: %d\n",therm_status.digital_readout);    
    printf("Tj: %d\n", therm_status.PROCHOT_temp);
    printf("Current Temp: %d\n", therm_status.current_core_temp);
}

void Read_Thermal_Status(struct therm_status_struct* therm_status) { 
    unsigned long val= get_msr_value(0,IA32_THERM_STATUS,63,0,&error_indx);
    therm_status->reading_valid = Get_Bits_Value(val,32,31);
    therm_status->resolution_deg_celcius = Get_Bits_Value(val,31,27);
    therm_status->digital_readout = Get_Bits_Value(val,23,16);
    
    val= get_msr_value(0,IA32_TEMPERATURE_TARGET,63,0,&error_indx);
    therm_status->PROCHOT_temp = Get_Bits_Value(val,24,16);
    
    therm_status->current_core_temp = therm_status->PROCHOT_temp - therm_status->digital_readout;
}

unsigned long Read_instruction_retired_any(){ return (get_msr_value(0,IA32_FIXED_CTR0,63,0,&error_indx));}
unsigned long Read_cpu_clk_unhalted_core(){ return (get_msr_value(0,IA32_FIXED_CTR1,63,0,&error_indx));}
unsigned long Read_cpu_clk_unhalted_ref(){ return (get_msr_value(0,IA32_FIXED_CTR2,63,0,&error_indx));}

//////////////PERFORMANCE COUNTER VERSION-1/////////////////////////////////////////////////////
void Set_MV1_Counters(int cpu,int pmc1, int pmc2, int pmc3, int pmc4){
    //pmc consists of 1byte of umask and 1byte of event select
    //63 = 0110 0011 = enable counter,anythread,os mode and user mode
    set_msr_value(cpu,IA32_PERFEVTSEL0,0x430000 + pmc1);
    set_msr_value(cpu,IA32_PERFEVTSEL1,0x430000 + pmc2);
    set_msr_value(cpu,IA32_PERFEVTSEL2,0x430000 + pmc3);
    set_msr_value(cpu,IA32_PERFEVTSEL3,0x430000 + pmc4);
}

unsigned long Get_MV1_Counter0(int cpu){return(get_msr_value(cpu,IA32_PMC0,63,0,&error_indx));}
unsigned long Get_MV1_Counter1(int cpu){return(get_msr_value(cpu,IA32_PMC1,63,0,&error_indx));}
unsigned long Get_MV1_Counter2(int cpu){return(get_msr_value(cpu,IA32_PMC2,63,0,&error_indx));}
unsigned long Get_MV1_Counter3(int cpu){return(get_msr_value(cpu,IA32_PMC3,63,0,&error_indx));}
///////////////////////////////////////////////////////////////////////////////////////////////


int Get_Bits_Value(unsigned long val,int highbit, int lowbit){ 
	unsigned long data = val;
	int bits = highbit - lowbit + 1;
	if(bits<64){
	    data >>= lowbit;
	    data &= (1ULL<<bits) - 1;
	}
	return(data);
}

struct cycle_information{
	unsigned long instruction_retired_any;
	unsigned long cpu_clk_unhalted_core;
	unsigned long cpu_clk_unhalted_ref;
	unsigned long mem_inst_retired_stores;
	unsigned long mem_inst_retired_loads;
	unsigned long fp_comp_ops_exe_sse_fp;
	unsigned long fp_comp_ops_exe_x87;
};

int main ()
{
	struct therm_status_struct therm_status;
	printf("Number of MSR %d\n",num_msr());
	printf("PMCx width %d\n",bit_width_PMCx());
	
	printf("Sizeof long %d\n",sizeof(long));
	printf("value of long %ld\n",0x7fffffffffffffff);
	
	unsigned long new_value=0, old_value=0;
	unsigned long new_value_array[4]={0,0,0,0}, old_value_array[4]={0,0,0,0};
	
	unsigned long instruction_retired_any_old=0, instruction_retired_any_new=0;
	unsigned long cpu_clk_unhalted_core_old=0, cpu_clk_unhalted_core_new=0;
	unsigned long cpu_clk_unhalted_ref_old=0, cpu_clk_unhalted_ref_new=0;
	unsigned long new_tsc=0, old_tsc=0;
	
/*	new_value = get_msr_value(0,0x186,63,0,&error_indx);
	printf("value %ld\n",new_value);
	new_value = get_msr_value(0,0x187,63,0,&error_indx);
	printf("value %ld\n",new_value);
	new_value = get_msr_value(0,0x188,63,0,&error_indx);
	printf("value %ld\n",new_value);*/
	
	
	//below is example of MV-1 for instruction_retired.any
	//set_msr_value(0,IA32_PERFEVTSEL0,0x63003c);
	Set_MV1_Counters(0,FP_COMP_OPS_EXE_SSE_FP,FP_COMP_OPS_EXE_X87,MEM_INST_RETIRED_LOADS,MEM_INST_RETIRED_STORES);

	//Needs to set this to enable Monitoring Version - 2
	set_msr_value(0,IA32_FIXED_CTR_CTL,0xFFF);		//can be used to stop/start counter
	set_msr_value(0,IA32_PERF_GLOBAL_CTRL,0x70000000FLLU);	//this controls both the MV-2 and MV-1 events
								//Thus need to enable it always!

	bool init_thread=1;

	while(1){
	    //old_value = new_value;
	    old_value_array[0] = new_value_array[0];old_value_array[1] = new_value_array[1];
	    old_value_array[2] = new_value_array[2];old_value_array[3] = new_value_array[3];
	    
	    old_tsc = new_tsc;
	    instruction_retired_any_old = instruction_retired_any_new;
	    cpu_clk_unhalted_core_old = cpu_clk_unhalted_core_new;
	    cpu_clk_unhalted_ref_old = cpu_clk_unhalted_ref_new;
	    
	    new_tsc = rdtsc();
	    //new_value = get_msr_value(0,IA32_PMC0,63,0,&error_indx);
	    new_value_array[0] = Get_MV1_Counter0(0);
	    new_value_array[1] = Get_MV1_Counter1(0);
	    new_value_array[2] = Get_MV1_Counter2(0);
	    new_value_array[3] = Get_MV1_Counter3(0);
	    
	    instruction_retired_any_new = Read_instruction_retired_any();
	    cpu_clk_unhalted_core_new = Read_cpu_clk_unhalted_core();
	    cpu_clk_unhalted_ref_new = Read_cpu_clk_unhalted_ref();

	    if(!init_thread){	    
		printf("\nCurrent time %d\n",time(NULL));
		printf("pmc value %ld\n",new_value);	    
		printf("tsc count %llu\t difference %ld\t ratio %f\n", (new_tsc - old_tsc), (cpu_clk_unhalted_core_new - cpu_clk_unhalted_core_old), ((double)(cpu_clk_unhalted_core_new - cpu_clk_unhalted_core_old))/((double)(new_tsc-old_tsc)));
		printf("Instruction retired any diff %lu\n", instruction_retired_any_new - instruction_retired_any_old);
		printf("CPU unhalted core diff %lu\n", cpu_clk_unhalted_core_new - cpu_clk_unhalted_core_old);
		printf("CPU unhalted ref diff %lu\n", cpu_clk_unhalted_ref_new - cpu_clk_unhalted_ref_old);

		printf("Mem Inst Retired Loads  %lu\n", new_value_array[0]-old_value_array[0]);
		printf("Mem Inst Retired Stores %lu\n", new_value_array[1]-old_value_array[1]);
		printf("FP_COMP_OPS_EXE.SSE_FP  %lu\n", new_value_array[2]-old_value_array[2]);
		printf("FP_COMP_OPS_EXE.X87     %lu\n", new_value_array[3]-old_value_array[3]);
	    }else{
		init_thread = false;
	    }
	    
	    Read_Thermal_Status(&therm_status);
	    Print_Thermal_Status(therm_status);
	    
	    sleep(1);
	}
	


        return(1);
}
